#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -*- Python -*-

"""
 @file RobotUSBCam.py
 @brief Periodically generate and save images and additional data
 @date $Date$

 @author East Japan Accounting Center Co., Ltd

 BSD 3-clause "New" or "Revised" License.

"""
import sys
import time
import datetime
import os
import math
from decimal import *
import ConfigParser
sys.path.append(".")

# Import RTM module
import RTC
import OpenRTM_aist

# Import OpenCV
import cv2

# ログ出力用
from logging import getLogger, StreamHandler, DEBUG
logger = getLogger(__name__)
handler = StreamHandler()
handler.setLevel(DEBUG)
logger.setLevel(DEBUG)
logger.addHandler(handler)

# 定数
# iniファイル名
INI_FILE = "sno.ini"

# iniファイルセクション名
INI_SECT = "last"

# iniファイル日付オプション名
INI_DAY = "day"

# iniファイルセクションNo名
INI_SNO = "sno"

# 画像保存先フォルダ名
IMAGE_FOLDER = "Picture/"

# 付加データ保存先フォルダ名
ADD_DATA_FOLDER = "Additional data/"

# 付加データの区切り文字
SEP = ","

# 画像の拡張子
IMAGE_EXT = ".jpg"

# 付加データの拡張子
ADD_DATA_EXT = ".csv"

# 撮影計画名文字数制限
MAX_LEN = 30

# Error Message
ERR_MSG_01 = "ERROR: USB camera is NOT available"
ERR_MSG_02 = "ERROR: Creation of save destination folder failed ( %(path)s )"
ERR_MSG_03 = "WARNING: Failed to save image"
ERR_MSG_04 = "WARNING: Failed to save csv data"

# Import Service implementation class
# <rtc-template block="service_impl">

# </rtc-template>

# Import Service stub modules
# <rtc-template block="consumer_import">
# </rtc-template>


# This module's spesification
# <rtc-template block="module_spec">
robotusbcam_spec = ["implementation_id", "RobotUSBCam",
                    "type_name",         "RobotUSBCam",
                    "description",       "Periodically generate and save images and additional data",
                    "version",           "1.0.0",
                    "vendor",            "EAC",
                    "category",          "Camera",
                    "activity_type",     "STATIC",
                    "max_instance",      "1",
                    "language",          "Python",
                    "lang_type",         "SCRIPT",
                    "conf.default.Interval", "1.0",
                    "conf.default.Width", "1280",
                    "conf.default.Height", "720",
                    "conf.default.RobotID", "drone1",
                    "conf.default.CameraNo", "1",
                    "conf.default.RootPath", "/home/pi/",
                    "conf.default.PlanName", "TestShot",
                    "conf.__widget__.Interval", "spin",
                    "conf.__widget__.Width", "spin",
                    "conf.__widget__.Height", "spin",
                    "conf.__widget__.RobotID", "text",
                    "conf.__widget__.CameraNo", "spin",
                    "conf.__widget__.RootPath", "text",
                    "conf.__widget__.PlanName", "text",
                    "conf.__constraints__.Interval", "x>=0.1",
                    "conf.__constraints__.Width", "x>=1",
                    "conf.__constraints__.Height", "x>=1",
                    ""]
# </rtc-template>

##
# @class RobotUSBCam
# @brief Periodically generate and save images and additional data
#
# Algorythm:  Refer to the "機能仕様書_RobotUSBCam" for details.
#
# In / Out:  Refer to the "機能仕様書_RobotUSBCam" for details.
#
# Overview:  Refer to the "機能仕様書_RobotUSBCam" for details.
#
#


class RobotUSBCam(OpenRTM_aist.DataFlowComponentBase):

    ##
    # @brief constructor
    # @param manager Maneger Object
    #
    def __init__(self, manager):
        OpenRTM_aist.DataFlowComponentBase.__init__(self, manager)

        self._d_inGPSDrone = RTC.GPSData(RTC.Time(0, 0), RTC.GPSTime(
            0, 0), 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0, RTC.GPS_FIX_NORMAL)
        """
        GPSデータ
         - Type: RTC::GPSData
         - Semantics: - Latitude
                      - Longitude
                      - Altitude
                      - TimeFromGPS
         - Unit: deg,sec
        """
        self._inGPSDroneIn = OpenRTM_aist.InPort(
            "inGPSRobot", self._d_inGPSDrone)
        self._d_inGyro = RTC.TimedOrientation3D(RTC.Time(0, 0), 0)
        """
        傾き情報
         - Type: RTC::TimedOrientation3D
         - Semantics: - Roll Axis
                      - Pitch Axis
                      - Yaw Axis
        """
        self._inGyroIn = OpenRTM_aist.InPort("inGyro", self._d_inGyro)

        # initialize of configuration-data.
        # <rtc-template block="init_conf_param">
        """
        画像キャプチャ間隔
         - Name: Interval Interval
         - DefaultValue: 1.0
         - Unit: sec
         - Range: 0.1<=x
        """
        self._Interval = [1.0]
        """
        撮影画像の幅
         - Name: Width Width
         - DefaultValue: 1280
         - Unit: px
         - Range: 1<=x
        """
        self._Width = [1280]
        """
        撮影画像の高さ
         - Name: Height Height
         - DefaultValue: 720
         - Unit: px
         - Range: 1<=x
        """
        self._Height = [720]
        """
        撮影ロボット識別子
         - Name: RobotID RobotID
         - DefaultValue: drone1
        """
        self._RobotID = ['drone1']
        """
        撮影カメラ識別番号
         - Name: CameraID CameraNo
         - DefaultValue: 1
        """
        self._CameraNo = [1]
        """
        画像および付加データ保存先ルートパス
         - Name: RootPath RootPath
         - DefaultValue: /home/pi/
        """
        self._RootPath = ['/home/pi/']
        """
        撮影計画名(半角30文字)
         - Name:  PlanName
         - DefaultValue: TestShot
        """
        self._PlanName = ['TestShot']

        # </rtc-template>

    ##
    # RTC初期化
    #
    # The initialize action (on CREATED->ALIVE transition)
    # formaer rtc_init_entry()
    #
    # @return RTC::ReturnCode_t
    #
    #
    def onInitialize(self):
        # Bind variables and configuration variable
        self.bindParameter("Interval", self._Interval, "1.0")
        self.bindParameter("Width", self._Width, "1280")
        self.bindParameter("Height", self._Height, "720")
        self.bindParameter("RobotID", self._RobotID, "drone1")
        self.bindParameter("CameraNo", self._CameraNo, "1")
        self.bindParameter("RootPath", self._RootPath, "/home/pi/")
        self.bindParameter("PlanName", self._PlanName, "TestShot")

        # Set InPort buffers
        self.addInPort("inGPSRobot", self._inGPSDroneIn)
        self.addInPort("inGyro", self._inGyroIn)

        # Set OutPort buffers

        # Set service provider to Ports

        # Set service consumers to Ports

        # Set CORBA Service Ports

        return RTC.RTC_OK

    #    ##
    #    #
    #    # The finalize action (on ALIVE->END transition)
    #    # formaer rtc_exiting_entry()
    #    #
    #    # @return RTC::ReturnCode_t
    #
    #    #
    # def onFinalize(self):
    #
    #    return RTC.RTC_OK

    #    ##
    #    #
    #    # The startup action when ExecutionContext startup
    #    # former rtc_starting_entry()
    #    #
    #    # @param ec_id target ExecutionContext Id
    #    #
    #    # @return RTC::ReturnCode_t
    #    #
    #    #
    # def onStartup(self, ec_id):
    #
    #    return RTC.RTC_OK

    #    ##
    #    #
    #    # The shutdown action when ExecutionContext stop
    #    # former rtc_stopping_entry()
    #    #
    #    # @param ec_id target ExecutionContext Id
    #    #
    #    # @return RTC::ReturnCode_t
    #    #
    #    #
    # def onShutdown(self, ec_id):
    #
    #    return RTC.RTC_OK

        ##
        # 変数初期化、カメラ情報取得、画像および付加データ保存先フォルダ作成
        #
        # The activated action (Active state entry action)
        # former rtc_active_entry()
        #
        # @param ec_id target ExecutionContext Id
        #
        # @return RTC::ReturnCode_t
        #
        #
    def onActivated(self, ec_id):

        # エラー判定用
        res = True

        # ロボットID
        self.robotID = self._RobotID[0]

        # 保存先フォルダルートパス
        self.rootPath = self._RootPath[0]

        # カメラNo
        self.camNo = self._CameraNo[0]

        # 撮影計画名 半角30文字を超えた場合30以下のみ使用
        self.planName = self.str_len_check(self._PlanName[0])

        # 日付取得
        self.today = datetime.date.today()

        # セクションNo
        self.sectionNo = self.get_section_no()

        # 撮影時刻保存用変数
        self.shotTime = self.get_now_time()

        # 画像フォルダ生成
        imagePath = self.create_image_path()
        try:
            os.makedirs(imagePath)
        except:
            res = False
            logger.error(ERR_MSG_02 % {"path": imagePath})

        # 画像フォルダ生成に失敗していなければ
        if res == True:
            # 付加データフォルダ生成
            adddataPath = self.create_adddata_path()
            try:
                os.makedirs(adddataPath)
            except:
                res = False
                logger.error(ERR_MSG_02 % {"path": adddataPath})

        # 失敗していなければ
        if res == True:
            # iniファイルのセクションNo+1
            self.set_section_no()

            # カメラ情報取得
            self.capture = cv2.VideoCapture(self.camNo - 1)

            # カメラ情報が取得できなければエラー
            if not self.capture.isOpened():
                res = False
                logger.error(ERR_MSG_01)

            # 画像の幅と高さ設定
            self.set_image_width_height()

        if res == True:
            return RTC.RTC_OK
        else:
            return RTC.RTC_ERROR

        ##
        # カメラ情報の解放
        #
        # The deactivated action (Active state exit action)
        # former rtc_active_exit()
        #
        # @param ec_id target ExecutionContext Id
        #
        # @return RTC::ReturnCode_t
        #
        #
    def onDeactivated(self, ec_id):

        # カメラ情報解放
        self.capture.release()

        return RTC.RTC_OK

        ##
        # USBカメラのデータ取得、画像および付加データの生成・保存
        #
        # The execution action that is invoked periodically
        # former rtc_active_do()
        #
        # @param ec_id target ExecutionContext Id
        #
        # @return RTC::ReturnCode_t
        #
        #
    def onExecute(self, ec_id):

        # GPS情報取得
        res = self.read_gps_drone()

        # inGPSDroneが取得できなかったら撮影しない
        if res:
            # 現在時刻取得
            nowTime = self.get_now_time()

            # 前回撮影時からInterval経過したら撮影
            timer = float(nowTime) - float(self.shotTime)
            if timer >= self._Interval[0]:

                # ジャイロ情報取得
                self.read_gyro()

                # カメラ画像キャプチャ取得　imageにデータが入る
                res, image = self.capture.read()

                # 時刻に変換
                self.setTime = datetime.datetime.fromtimestamp(
                    float(nowTime))

                # 画像名設定
                imageName = self.create_file_name() + IMAGE_EXT

                # 付加データ名設定
                addDataName = self.create_file_name() + ADD_DATA_EXT

                # 付加データ作成
                addData = self.create_adddata()

                # 画像保存
                try:
                    res = cv2.imwrite(
                        self.create_image_path() + imageName, image)
                    if res == False:
                        raise Exception()
                except:
                    logger.warning(ERR_MSG_03)
                    res = False

                # 画像保存失敗時は保存しない
                if res == True:
                    # 付加データ保存
                    try:
                        with open(self.create_adddata_path() + addDataName, "w") as createText:
                            createText.write(addData)
                    except:
                        logger.warning(ERR_MSG_04)

                self.shotTime = nowTime

        return RTC.RTC_OK

    #    ##
    #    #
    #    # The aborting action when main logic error occurred.
    #    # former rtc_aborting_entry()
    #    #
    #    # @param ec_id target ExecutionContext Id
    #    #
    #    # @return RTC::ReturnCode_t
    #    #
    #    #
    # def onAborting(self, ec_id):
    #
    #    return RTC.RTC_OK

    #    ##
    #    #
    #    # The error action in ERROR state
    #    # former rtc_error_do()
    #    #
    #    # @param ec_id target ExecutionContext Id
    #    #
    #    # @return RTC::ReturnCode_t
    #    #
    #    #
    # def onError(self, ec_id):
    #
    #    return RTC.RTC_OK

    #    ##
    #    #
    #    # The reset action that is invoked resetting
    #    # This is same but different the former rtc_init_entry()
    #    #
    #    # @param ec_id target ExecutionContext Id
    #    #
    #    # @return RTC::ReturnCode_t
    #    #
    #    #
    # def onReset(self, ec_id):
    #
    #    return RTC.RTC_OK

    #    ##
    #    #
    #    # The state update action that is invoked after onExecute() action
    #    # no corresponding operation exists in OpenRTm-aist-0.2.0
    #    #
    #    # @param ec_id target ExecutionContext Id
    #    #
    #    # @return RTC::ReturnCode_t
    #    #

    #    #
    # def onStateUpdate(self, ec_id):
    #
    #    return RTC.RTC_OK

    #    ##
    #    #
    #    # The action that is invoked when execution context's rate is changed
    #    # no corresponding operation exists in OpenRTm-aist-0.2.0
    #    #
    #    # @param ec_id target ExecutionContext Id
    #    #
    #    # @return RTC::ReturnCode_t
    #    #
    #    #
    # def onRateChanged(self, ec_id):
    #
    #    return RTC.RTC_OK

    # 現在時刻取得
    def get_now_time(self):
        return "%0.3f" % time.time()

    # iniファイルからセクションNo取得
    def get_section_no(self):
        # iniファイル取得
        inifile = ConfigParser.SafeConfigParser()
        inifile.read(INI_FILE)

        # 前回撮影日
        lastDay = inifile.get(INI_SECT, INI_DAY)

        # 前回撮影日と違ったらセクションNoを1に
        if lastDay != str(self.today):
            inifile.set(INI_SECT, INI_SNO, str(1))
            inifile.set(INI_SECT, INI_DAY, str(self.today))
            with open("./" + INI_FILE, "w") as createIni:
                inifile.write(createIni)
            inifile.read(INI_FILE)

        return inifile.getint(INI_SECT, INI_SNO)

    # iniファイルにセクションNo+1
    def set_section_no(self):
        inifile = ConfigParser.SafeConfigParser()
        inifile.read(INI_FILE)
        inifile.set(INI_SECT, INI_SNO, str(self.sectionNo + 1))
        with open("./" + INI_FILE, "w") as createIni:
            inifile.write(createIni)

    # 画像の幅と高さ設定
    def set_image_width_height(self):
        self.capture.set(cv2.cv.CV_CAP_PROP_FRAME_WIDTH, self._Width[0])
        self.capture.set(cv2.cv.CV_CAP_PROP_FRAME_HEIGHT, self._Height[0])

    # 現在位置取得
    def read_gps_drone(self):
        if self._inGPSDroneIn.isNew():
            gpsDrone = self._inGPSDroneIn.read()
            self.lat = gpsDrone.latitude
            self.lon = gpsDrone.longitude
            self.alt = gpsDrone.altitude
            ret = True
        else:
            ret = False

        return ret

    # ジャイロ情報取得
    def read_gyro(self):
        if self._inGyroIn.isNew():
            gyro = self._inGyroIn.read()
            self.roll = gyro.data.r
            self.pitch = gyro.data.p
            self.yaw = gyro.data.y
        else:
            self.roll = 0.0
            self.pitch = 0.0
            self.yaw = 0.0

    # フォルダパス生成
    def create_folder_path(self):
        return self.robotID + self.today.strftime("_%Y%m%d_") + str(self.sectionNo)

    # 画像パス生成
    def create_image_path(self):
        return self.rootPath + IMAGE_FOLDER + self.create_folder_path() + "/"

    # 付加データパス生成
    def create_adddata_path(self):
        return self.rootPath + ADD_DATA_FOLDER + self.create_folder_path() + "/"

    # ファイル名生成
    def create_file_name(self):
        ret = self.today.strftime("%Y%m%d_") + \
            self.setTime.strftime("%H%M%S") + "%03d" % (self.setTime.microsecond // 1000) + "_" + \
            self.robotID + "_" + str(self.camNo)

        return ret

    # 付加データ生成
    def create_adddata(self):
        ret = self.create_file_name() + IMAGE_EXT + SEP + \
            self.create_folder_path() + SEP + \
            str(self.capture.get(cv2.cv.CV_CAP_PROP_FRAME_WIDTH)) + SEP + \
            str(self.capture.get(cv2.cv.CV_CAP_PROP_FRAME_HEIGHT)) + SEP + \
            self.robotID + SEP + \
            str(self.camNo) + SEP + \
            str(self.lat_range_check(self.ceiling_digit(self.lat))) + SEP + \
            str(self.range_check(self.ceiling_digit(self.lon))) + SEP + \
            str(self.ceiling_digit(self.alt)) + SEP + \
            str(self.rad_to_deg_yaw(self.yaw)) + SEP + \
            str(self.range_check(math.degrees(self.pitch))) + SEP + \
            str(self.range_check(math.degrees(self.roll))) + SEP + \
            self.today.strftime("%Y%m%d") + SEP + \
            self.setTime.strftime("%H%M%S") + "%03d" % (self.setTime.microsecond // 1000) + SEP + \
            str(self.planName)

        return ret

    # 少数点6桁で切り上げ
    def ceiling_digit(self, data):
        ret = Decimal(data).quantize(
            Decimal(".000000"), rounding=ROUND_CEILING)
        return ret

    # rangeを90°以下になるようにする
    def lat_range_check(self, lat):
        if lat > 90.0:
            lat = 90.0
        elif lat < -90.0:
            lat = -90.0

        return lat

    # rangeを-180~180°になるようにする
    def range_check(self, range):
        if range > 180.0:
            ranRem = range % 360.0
            if ranRem > 180.0:
                range = ranRem - 360.0
            else:
                range = ranRem
        elif range < -180.0:
            ranRem = abs(range) % 360.0
            if ranRem > 180.0:
                range = 360.0 - ranRem
            else:
                range = -ranRem

        return range

    # radを0~360°に変換
    def rad_to_deg_yaw(self, rad):
        ret = math.degrees(rad)
        if ret > 360:
            ret = ret % 360
        elif ret < 0:
            ret = 360 - (abs(ret) % 360)

        return ret

    # unicode文字をMAX_LEN文字以下にする
    def str_len_check(self, str):
        str = str.decode("UTF-8").encode("shift-jis")
        if len(str) > MAX_LEN:
            str = str[:MAX_LEN]
        str = str.decode("shift-jis").encode("UTF-8")

        return str


def RobotUSBCamInit(manager):
    profile = OpenRTM_aist.Properties(defaults_str=robotusbcam_spec)
    manager.registerFactory(profile,
                            RobotUSBCam,
                            OpenRTM_aist.Delete)


def MyModuleInit(manager):
    RobotUSBCamInit(manager)

    # Create a component
    comp = manager.createComponent("RobotUSBCam")


def main():
    mgr = OpenRTM_aist.Manager.init(sys.argv)
    mgr.setModuleInitProc(MyModuleInit)
    mgr.activateManager()
    mgr.runManager()

if __name__ == "__main__":
    main()
